package org.colllib.builder;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.colllib.datastruct.Pair;

/**
 * Helper class to create a filled map on the fly. Basically, you call
 * <code>create</code> to set up a builder, <code>put</code> to fill the map,
 * <code>build</code> to retrieve the result:
 * <pre>
 *    System.out.println(
 *      MapBuilder.&lt;Integer, String&gt;createHash()
 *        .put(1, "a")
 *        .put(2, "b")
 *        .build());
 *    ==> {1=a, 2=b}
 * </pre>
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 *
 * @author marc.jackisch
 */
public class MapBuilder<K, V, M extends Map<K, V>> {
   /**
    * Create a new map builder for the given map
    * 
    * @param <K>  key type
    * @param <V>  value type
    * @param <M>  map type
    * @param m the map
    * @return a map builder for m
    */
   public static <K, V, M extends Map<K, V>> MapBuilder<K, V, M> create(M m) {
      return new MapBuilder<K, V, M>(m);
   }
   
   /**
    * Create a new map builder for a new hash map
    * 
    * @param <K>  key type
    * @param <V>  value type
    * @return a map builder for a new hash map
    */
   public static <K, V> MapBuilder<K, V, HashMap<K, V>> createHash() {
      return create(new HashMap<K, V>());
   }

   private final M m;

   private MapBuilder(M m) {
      this.m = m;
   }

   /**
    * Add an entry to the underlying map
    * 
    * @param k  key
    * @param v  value
    * @return this
    */
   public MapBuilder<K, V, M> put(K k, V v) {
      m.put(k, v);
      return this;
   }

   /**
    * Add all map entries to the underlying mao
    * 
    * @param map the map containing the entries to add
    * @return this
    */
   public MapBuilder<K, V, M> put(Map<? extends K, ? extends V> map) {
      for(Entry<? extends K, ? extends V> e : map.entrySet())
         m.put(e.getKey(), e.getValue());
      return this;
   }

   /**
    * Add key/value pairs to the underlying map
    * 
    * @param keyValPairs the keys and values as pairs
    * @return this
    */
   public MapBuilder<K, V, M> put(Iterable<Pair<? extends K, ? extends V>> keyValPairs) {
      for(Pair<? extends K, ? extends V> p : keyValPairs)
         m.put(p.x, p.y);
      return this;
   }

   /**
    * Fetch the underlying map
    * 
    * @return the underlying map of this builder
    */
   public M build() {
      return m;
   }
}
