package org.colllib.caches;

import java.util.HashMap;

import org.colllib.introspect.Introspector;
import org.colllib.introspect.PropertyAccessor;

/**
 * Basically the same as read- and write method cache, this class stores a
 * property accessor determined via introspection for a concrete class.
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0. See
 * the file licenses/lgpl-3.0.txt for details.
 * 
 * @author mjackisch
 */
public class PropertyAccessorCache<V> {
   private final String                        property;
   private HashMap<Class<?>, PropertyAccessor> propertyAccessors;

   public PropertyAccessorCache(String property) {
      this.property = property;
   }

   public PropertyAccessor getPropertyAccessor(Class<?> cl) {
      if(propertyAccessors == null)
         propertyAccessors = new HashMap<Class<?>, PropertyAccessor>();
      PropertyAccessor m = propertyAccessors.get(cl);
      if(m != null)
         return m;
      try {
         for(PropertyAccessor pa : Introspector.introspect(cl))
            if(pa.getName().equals(property)) {
               propertyAccessors.put(cl, pa);
               return pa;
            }
         throw new RuntimeException("No accessor for property \""
            + property + "\" found in class " + cl.getCanonicalName());
      }
      catch(Exception e) {
         throw new RuntimeException(e);
      }
   }

   public String getProperty() {
      return property;
   }

   @SuppressWarnings("unchecked")
   public V readValue(Object t) {
      try {
         Object v = getPropertyAccessor(t.getClass()).getPropertyValue(t);
         return (V) v;
      }
      catch(Exception e) {
         throw new RuntimeException(e);
      }
   }
   
   public void writeValue(Object t, Object v) {
      try {
         PropertyAccessor pa = getPropertyAccessor(t.getClass());
         if(!pa.isWritable())
            throw new RuntimeException("Property \"" + property + "\" is not "
               + "writable for class " + t.getClass().getCanonicalName());
         pa.setPropertyValue(t, v);
      }
      catch(Exception e) {
         throw new RuntimeException(e);
      }
   }
}
