package org.colllib.datastruct;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.colllib.transformer.Transformer;
import org.colllib.transformer.TransformerCollection;
import org.colllib.util.CollectionUtil;

/**
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marcel.billen
 */
public class AutoRemoverMap<K, V, D> implements Map<K, V>, Serializable {
   public static interface RemoveDecider<K, V, D> extends Serializable {
      public D createCriteria(K key, V value);
      public boolean shallRemove(K key, V value, D criteria);
   }
   
   private Map<K, Pair<V, D>> backingMap;
   private RemoveDecider<K, V, D> decider;

   public AutoRemoverMap(RemoveDecider<K, V, D> decider) {
      this.backingMap = new HashMap<K, Pair<V,D>>();
      this.decider = decider;
   }

   @Override
   public int size() {
      removeElements();
      return backingMap.size();
   }

   @Override
   public boolean isEmpty() {
      removeElements();
      return backingMap.isEmpty();
   }

   @Override
   public boolean containsKey(Object key) {
      removeElements();
      return backingMap.containsKey(key);
   }

   @Override
   public boolean containsValue(Object value) {
      removeElements();
      return CollectionUtil.transform(
         backingMap.values(), 
         TransformerCollection.<Pair<V, D>, V>beanValue("x")).contains(value);
   }

   @Override
   public V get(Object key) {
      removeElements();
      Pair<V, D> val = backingMap.get(key);
      return val != null ? val.x : null;
   }

   @Override
   public V put(K key, V value) {
      removeElements();
      Pair<V, D> prev = backingMap.put(key, new Pair<V, D>(value, decider.createCriteria(key, value)));
      return prev != null ? prev.x : null;
   }

   @Override
   public V remove(Object key) {
      removeElements();
      Pair<V, D> prev = backingMap.remove(key);
      return prev != null ? prev.x : null;
   }

   @Override
   public void putAll(Map<? extends K, ? extends V> m) {
      removeElements();
      for(Map.Entry<? extends K, ? extends V> entry : m.entrySet()) 
         put(entry.getKey(), entry.getValue());
   }

   @Override
   public void clear() {
      backingMap.clear();
   }

   @Override
   public Set<K> keySet() {
      return backingMap.keySet();
   }

   @Override
   public Collection<V> values() {
      return CollectionUtil.transform(backingMap.values(), TransformerCollection.<Pair<V, D>, V>beanValue("x"));
   }

   @Override
   public Set<Map.Entry<K, V>> entrySet() {
      return CollectionUtil.transformToSet(
         backingMap.entrySet(),
         new EntryTransformer<K, V, D>(decider));
   }

   @Override
   public boolean equals(Object o) {
      return backingMap.equals(o);
   }

   @Override
   public int hashCode() {
      return backingMap.hashCode();
   }
   
   private void removeElements() {
      ArrayList<K> keysToRemove = new ArrayList<K>();
      for(Map.Entry<K,Pair<V,D>> entry : backingMap.entrySet()) {
         K key = entry.getKey();
         Pair<V,D> value = entry.getValue();
         if(decider.shallRemove(key, value.x, value.y))
            keysToRemove.add(key);
      } 
      CollectionUtil.removeAllKeys(backingMap, keysToRemove);
   }
   
   private static class EntryTransformer<K, V, D> implements Transformer<Map.Entry<K, Pair<V, D>>, Map.Entry<K, V>>, Serializable {
      private RemoveDecider<K, V, D> decider;
      
      public EntryTransformer(RemoveDecider<K, V, D> decider) {
         this.decider = decider;
      }

      @Override
      public java.util.Map.Entry<K, V> transform(java.util.Map.Entry<K, Pair<V, D>> backingEntry) {
         return new WrapperEntry<K, V, D>(backingEntry, decider);
      }
   }
   
   private static class WrapperEntry<K, V, D> implements Map.Entry<K, V>, Serializable {
      private final Map.Entry<K, Pair<V, D>> backingEntry;
      private RemoveDecider<K, V, D> decider;
      
      public WrapperEntry(Map.Entry<K, Pair<V, D>> backingEntry, RemoveDecider<K, V, D> decider) {
         this.backingEntry = backingEntry;
         this.decider = decider;
      }

      @Override
      public K getKey() {
         return backingEntry.getKey();
      }

      @Override
      public V getValue() {
         return backingEntry.getValue().x;
      }

      @Override
      public V setValue(V value) {
         Pair<V, D> curr = backingEntry.getValue();
         V old = curr.x;
         curr.x = value;
         curr.y = decider.createCriteria(getKey(), value);
         return old;
      }

      @Override
      public boolean equals(Object o) {
         return backingEntry.equals(o);
      }

      @Override
      public int hashCode() {
         return backingEntry.hashCode();
      }
   }
}

