package org.colllib.datastruct;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import org.colllib.util.CompareUtil;

/**
 * Simple typed tree implementation with ordered child lists.
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marc.jackisch
 */
public class Tree<T> implements Serializable {
   private ArrayList<Tree<T>> children;
   private Tree<T>            parent;
   private T                  data;
   
   /**
    * Create a new tree node with a given parent and payload
    * 
    * @param parent  the parent node, or null for a root
    * @param data    the user data object
    */
   private Tree(Tree<T> parent, T data) {
      this.parent = parent;
      this.data = data;
      this.children = new ArrayList<Tree<T>>();
   }
   
   /**
    * Return the child tree nodes
    * 
    * @return ordered list of children
    */
   public List<Tree<T>> getChildren() {
      return Collections.unmodifiableList(children);
   }
   
   /**
    * @return the parent node
    */
   public Tree<T> getParent() {
      return parent;
   }
   
   /**
    * Set the parent node for this node
    * 
    * @param parent new parent node
    */
   public void setParent(Tree<T> parent) {
      this.parent = parent;
   }
   
   /**
    * @return the user data object
    */
   public T getData() {
      return data;
   }
   
   /**
    * Set the user data object
    * 
    * @param data new data object
    */
   public void setData(T data) {
      this.data = data;
   }
   
   /**
    * Add a child to the list of child nodes
    * 
    * @param child the new child
    */
   public void addChild(Tree<T> child) {
      children.add(child);
      child.parent = this;
   }
   
   /**
    * Add a new child tree node for a given data object
    * 
    * @param data the new child data object
    */
   public void addChild(T data) {
      addChild(new Tree<T>(this, data));
   }

   /**
    * Insert a child node at a given position of the child list
    * 
    * @param index the index
    * @param child the child node
    */
   public void insertChild(int index, Tree<T> child) {
      children.add(index, child);
      child.parent = this;
   }

   /**
    * Insert a data object at a given position of the child list
    * 
    * @param index the index
    * @param data the new child data object
    */
   public void insertChild(int index, T data) {
      insertChild(index, new Tree<T>(this, data));
   }
   
   /**
    * Remove a given child node from the child list
    * 
    * @param child the child node
    * @return <code>true</code> if the node was in the child list and has been
    * removed, <code>false</code> if nothing happened
    */
   public boolean removeChild(Tree<T> child) {
      boolean res = children.remove(child);
      if(res)
         child.parent = null;
      return res;
   }
   
   /**
    * Remove the given user data object from the child list (the first occurance
    * in case there are multiple copies)
    * 
    * @param data  the data object to remove
    * @return <code>true</code> if the object has been found and removed,
    * <code>false</code> if not
    */
   public boolean removeChild(T data) {
      for(Tree<T> c : children)
         if(CompareUtil.nullSafeEquals(c.data, data))
            return removeChild(c);
      return false;
   }

   /**
    * @return the number of children
    */
   public int getChildCount() {
      return children.size();
   }

   /**
    * @return true, if the node does not have a parent
    */
   public boolean isRoot() {
      return parent == null;
   }
   
   /**
    * @return true, if the node does not have children
    */
   public boolean isLeaf() {
      return children.isEmpty();
   }
   
   /**
    * Determines this node's path from root
    * 
    * @return a list of tree nodes, so that the tree root is at index 0, and
    * <code>this</code> is the last list element 
    */
   public List<Tree<T>> pathFromRoot() {
      LinkedList<Tree<T>> res = new LinkedList<Tree<T>>();
      Tree<T> t = this;
      do {
         res.add(0, t);
         t = t.getParent();
      }
      while(t != null);
      return res;
   }
   
   /**
    * Return a breadth first list projection of the tree's data
    * 
    * @return a list of data objects
    */
   public List<T> breadthFirst() {
      ArrayList<T> res = new ArrayList<T>();
      ArrayList<Tree<T>> q = new ArrayList<Tree<T>>();
      q.add(this);
      while(!q.isEmpty()) {
         Tree<T> t = q.remove(0);
         res.add(t.data);
         q.addAll(t.children);
      }
      return res;
   }
   
   /**
    * Return a breadth first list projection of the tree nodes
    * 
    * @return a list of tree nodes
    */
   public List<Tree<T>> breadthFirstNodes() {
      ArrayList<Tree<T>> res = new ArrayList<Tree<T>>();
      ArrayList<Tree<T>> q = new ArrayList<Tree<T>>();
      q.add(this);
      while(!q.isEmpty()) {
         Tree<T> t = q.remove(0);
         res.add(t);
         q.addAll(t.children);
      }
      return res;
   }
   
   /**
    * Return a depth first list projection of the tree's data
    * 
    * @return a list of data objects
    */
   public List<T> depthFirst() {
      ArrayList<T> res = new ArrayList<T>();
      Stack<Tree<T>> s = new Stack<Tree<T>>();
      s.push(this);
      while(!s.isEmpty()) {
         Tree<T> t = s.pop();
         res.add(t.data);
         for(int i = t.children.size() - 1; i >= 0; i--)
            s.push(t.children.get(i));
      }
      return res;
   }
   
   /**
    * Return a depth first list projection of the tree nodes
    * 
    * @return a list of tree nodes
    */
   public List<Tree<T>> depthFirstNodes() {
      ArrayList<Tree<T>> res = new ArrayList<Tree<T>>();
      Stack<Tree<T>> s = new Stack<Tree<T>>();
      s.push(this);
      while(!s.isEmpty()) {
         Tree<T> t = s.pop();
         res.add(t);
         for(int i = t.children.size() - 1; i >= 0; i--)
            s.push(t.children.get(i));
      }
      return res;
   }
}
