package org.colllib.filter;

import java.io.Serializable;
import java.util.regex.Pattern;

import org.colllib.caches.PropertyAccessorCache;
import org.colllib.transformer.Transformer;

/**
 * This class is a factory for various filters.
 * 
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 *
 * @author marc.jackisch
 */
public class FilterCollection {
   private FilterCollection() {};

   private static class BeanValueEqualsFilter<T, V> implements Filter<T>, Serializable {
      private final PropertyAccessorCache<V> pac;
      private final V                        comparisonValue;

      public BeanValueEqualsFilter(String beanValue, V comparisonValue) {
         super();
         this.comparisonValue = comparisonValue;

         pac = new PropertyAccessorCache<V>(beanValue);
      }

      @Override
      public boolean matches(T t) {
         if(t == null)
            return false;

         V v = pac.readValue(t);
         if(v == null && comparisonValue == null)
            return true;
         if(comparisonValue != null && v == null)
            return false;
         if(comparisonValue == null && v != null)
            return false;
         if(comparisonValue == v)
            return true;
         return comparisonValue.equals(v);
      }
   }

   /**
    * Create a generic reflection based filter, which evaluates to true, if
    * a given property on the input bean (accessible via a bean standard getter,
    * i.e. getPropname or isPropname) matches a given comparison value.
    * 
    * @param <T>              the bean type
    * @param <V>              the property type
    * @param beanValue        the property name
    * @param comparisonValue  the match value
    * @return a filter for type T
    */
   public static <T, V> Filter<T> beanValueEqualsFilter(String beanValue, V comparisonValue) {
      return new BeanValueEqualsFilter<T, V>(beanValue, comparisonValue);
   }


   private static class Not<T> implements Filter<T>, Serializable {
      private Filter<T> reference;

      public Not(Filter<T> reference) {
         this.reference = reference;
      }

      @Override
      public boolean matches(T t) {
         return !reference.matches(t);
      }
   }


   /**
    * Create a filter that negates another filter
    * 
    * @param <T>   the base type
    * @param f     the underlying filter
    * @return      a filter, wich performs a "not f" evaluation
    */
   public static <T> Filter<T> not(Filter<T> f) {
      return new Not<T>(f);
   }


   private static class And<T> implements Filter<T>, Serializable {
      private Filter<T>[] f;

      public And(Filter<T> ... f) {
         this.f = f;
      }

      @Override
      public boolean matches(T t) {
         for(Filter<T> fi : f)
            if(fi != null && !fi.matches(t))
               return false;
         return true;
      }
   }

   /**
    * Create a filter that performs a logical "and" operation on a list of
    * other filters
    * @param <T>   the base type
    * @param f     one or more filters
    * @return a filter, that evaluates to true, if all underlying filters
    * evaluate to true
    */
   public static <T> Filter<T> and(Filter<T> ... f) {
      return new And<T>(f);
   }


   private static class Or<T> implements Filter<T>, Serializable {
      private Filter<T>[] f;

      public Or(Filter<T> ... f) {
         this.f = f;
      }

      @Override
      public boolean matches(T t) {
         for(Filter<T> fi : f)
            if(fi != null && fi.matches(t))
               return true;
         return false;
      }
   }

   /**
    * Create a filter that performs a logical "or" operation on a list of
    * other filters
    * @param <T>   the base type
    * @param f     one or more filters
    * @return a filter, that evaluates to true, if at least one underlying filter
    * evaluates to true
    */
   public static <T> Filter<T> or(Filter<T> ... f) {
      return new Or<T>(f);
   }


   private static class Yes<T> implements Filter<T>, Serializable {
      @Override
      public boolean matches(T t) {
         return true;
      }
   };

   /**
    * Return an "always true" filter
    * 
    * @param <T> the base type
    * @return a filter that returns true
    */
   public static <T> Filter<T> yes() {
      return new Yes<T>();
   }


   /**
    * Return a filter, which performs a regular expression match on the string
    * representation of the objects to be evaluated
    * 
    * @param <T>     the base type
    * @param match   regexp to check against
    * @return filter, that evaluates to true for an object x, if
    * match.matches(x.toString()) (basically)
    */
   public static <T> Filter<T> matches(String match) {
      return new StringMatchFilter<T>(match);
   }

   private static class StringMatchFilter<T> implements Filter<T>, Serializable {
      private Pattern pat;

      public StringMatchFilter(String s) {
         pat = Pattern.compile(s);
      }

      @Override
      public boolean matches(T t) {
         if(t == null)
            return false;
         return pat.matcher(t.toString()).matches();
      }
   }


   /**
    * Return a filter, that checks a transformation result of the base object
    * against another given filter.
    * 
    * @param <T>          the base type
    * @param <S>          the transformation result type
    * @param transformer  the transformer to use
    * @param filter       the filter to check
    * @return a filter, that basically evaluates to
    * filter.matches(transformer.transform(inputObject))
    */
   public static <T, S> Filter<T> transformMatches(Transformer<T, S> transformer, Filter<S> filter) {
      return new TransformMatchesFilter<T, S>(transformer, filter);
   }

   private static class TransformMatchesFilter<T, S> implements Filter<T>, Serializable {
      private Transformer<T, S>    transformer;
      private Filter<S>            filter;

      public TransformMatchesFilter(Transformer<T, S> transformer, Filter<S> filter) {
         this.transformer = transformer;
         this.filter = filter;
      }

      @Override
      public boolean matches(T t) {
         return filter.matches(transformer.transform(t));
      }
   }


   /**
    * Return a filter that checks, whether the objects can be assigned to a
    * given class.
    * 
    * @param <T>  the base class
    * @param <C>  the destination class
    * @param c Class<C>
    * @return a filter, that checks whether the input objects can be assigned to
    * a pointer of type c.
    */
   public static <T, C> Filter<T> isAssignableFromFilter(Class<C> c) {
      return new IsAssignableFromFilter<T, C>(c);
   }

   private static class IsAssignableFromFilter<T, C> implements Filter<T>, Serializable {
      private Class<C>  c;

      public IsAssignableFromFilter(Class<C> c) {
         this.c = c;
      }

      @Override
      public boolean matches(T t) {
         return t == null || c.isAssignableFrom(t.getClass());
      }
   }
   
   
   /**
    * Return a filter that matches a given element in terms of an equals-call
    * 
    * @param <T>   the filter class
    * @param val   the value to match
    * @return a filter, that matches for val == null && element == null or
    * val.equals(element)
    */
   public static <T> Filter<T> equalsFilter(T val) {
      return new EqualsFilter<T>(val);
   }
   
   private static class EqualsFilter<T> implements Filter<T>, Serializable {
      private T val;

      public EqualsFilter(T val) {
         this.val = val;
      }

      @Override
      public boolean matches(T t) {
         return (val == null && t == null)
           || (val != null && t != null && val.equals(t));
      }
   }

   /**
    * Filter that matches null
    */
   public static <T> Filter<T> nullFilter() {
      return equalsFilter(null);
   }
}
