package org.colllib.introspect;

import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.colllib.caches.GenCache;
import org.colllib.datastruct.AutoInitMap;
import org.colllib.filter.Filter;
import org.colllib.transformer.TransformerCollection;
import org.colllib.util.CollectionUtil;

/**
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author mjackisch
 */
public class Introspector {
   private static final Pattern MPAT = Pattern.compile("(get|is|set)(.*)");

   private static GenCache<Class<?>, List<PropertyAccessor>> CACHE
   = new GenCache<Class<?>, List<PropertyAccessor>>(new GenCache.LookupProvider<Class<?>, List<PropertyAccessor>>() {
      @Override
      public List<PropertyAccessor> lookup(Class<?> k) {
         return introspectInternal(k);
      }});
   
   public static List<PropertyAccessor> introspect(Class<?> tClass) {
      return CACHE.get(tClass);
   }
   
   private static List<PropertyAccessor> introspectInternal(Class<?> tClass) {
      AutoInitMap<String, PropertyAccessor> mcoll
      = new AutoInitMap<String, PropertyAccessor>(
         new HashMap<String, PropertyAccessor>(),
         TransformerCollection.constructorCall(String.class, PropertyAccessor.class)); 

      try {
         for(PropertyDescriptor pd : java.beans.Introspector.getBeanInfo(tClass).getPropertyDescriptors()) {
            if(pd.getReadMethod() != null) {
               mcoll.get(pd.getName()).setReadMethod(pd.getReadMethod());
               mcoll.get(pd.getName()).setWriteMethod(pd.getWriteMethod());
            }
         }
      }
      catch(IntrospectionException e) {
         throw new RuntimeException(e);
      }
      HashSet<String> beanInfoProps = new HashSet<String>(mcoll.keySet());
      
      for(Method m : tClass.getMethods()) {
         if(Modifier.isPublic(m.getModifiers()) && !Modifier.isStatic(m.getModifiers())) {
            Matcher matcher = MPAT.matcher(m.getName());
            if(matcher.matches()) {
               String propName = matcher.group(2);
               if(propName.length() == 0)
                  continue;
               
               propName = propName.substring(0, 1).toLowerCase() + propName.substring(1);
               if(beanInfoProps.contains(propName))
                  continue;
               
               String prefix = matcher.group(1);
               boolean isWriteMethod = prefix.equals("set");
               if(isWriteMethod && m.getParameterTypes().length != 1)
                  continue;
               if(!isWriteMethod && m.getParameterTypes().length != 0)
                  continue;
               if(!isWriteMethod && m.getReturnType().equals(Void.TYPE))
                  continue;
               
               if(isWriteMethod)
                  mcoll.get(propName).setWriteMethod(m);
               else
                  mcoll.get(propName).setReadMethod(m);
            }
         }
      }
      
      for(Field f : tClass.getFields()) {
         if(Modifier.isPublic(f.getModifiers()) && !Modifier.isStatic(f.getModifiers()))
            mcoll.get(f.getName()).setPublicField(f);
      }
      
      ArrayList<PropertyAccessor> res = CollectionUtil.applyFilter(
         mcoll.values(),
         new Filter<PropertyAccessor>() {
            @Override
            public boolean matches(PropertyAccessor t) {
               return t.getReadMethod() != null || t.getPublicField() != null;
            }});
      
      return Collections.unmodifiableList(res);
   }
   
   
   public static interface SupI {
      public String getSupi();
      public void setSupi(String s);
   }

   public static interface SubI extends SupI {
      public String getSubi();
      public void setSubi(String s);
   }

   public static abstract class TotalDrueber {
      public String totalTopLevelJawoll;
   }
   
   public static class Ueberdings extends TotalDrueber {
      private String ueberdings;
      public String ueberpubliza;

      public String getUeberdings() {
         return ueberdings;
      }

      public void setUeberdings(String ueberdings) {
         this.ueberdings = ueberdings;
      }
   }
   
   public static class Dingsda extends Ueberdings implements SubI {
      private String supi;
      private String subi;
      private String dings;
      public String publiza;
      
      @Override
      public String getSupi() {
         return supi;
      }
      @Override
      public void setSupi(String supi) {
         this.supi = supi;
      }
      @Override
      public String getSubi() {
         return subi;
      }
      @Override
      public void setSubi(String subi) {
         this.subi = subi;
      }
      public String getDings() {
         return dings;
      }
      public void setDings(String dings) {
         this.dings = dings;
      }
   }
   
   public static void main(String[] args) {
//      System.out.println(introspect(ArrayList.class));
//      System.out.println("-----");
//      System.out.println(introspect(Pair.class));
      System.out.println(introspect(Dingsda.class));
      System.out.println(introspect(SubI.class));
      
      Dingsda d = new Dingsda();
      List<PropertyAccessor> pds = introspect(Dingsda.class);
      for(PropertyAccessor pa : pds)
         if(pa.isWritable())
            pa.setPropertyValue(d, pa.getName() + "_Value");
      for(PropertyAccessor pa : pds)
         System.out.println(pa.getName() + ": " + pa.getPropertyValue(d));
      
      int n = 1000000;
      long st1 = System.currentTimeMillis();
      for(int i = 0; i < n; i++) {
         try {
            java.beans.Introspector.getBeanInfo(Dingsda.class).getPropertyDescriptors();
         }
         catch(IntrospectionException e) {
            e.printStackTrace();
         }
      }
      long st2 = System.currentTimeMillis();
      for(int i = 0; i < n; i++)
         introspect(Dingsda.class);
      long st3 = System.currentTimeMillis();
      System.out.println("BI: " + (st2 - st1));
      System.out.println("IN: " + (st3 - st2));
   }
   
   
   // Momentn! Wie schaut's aus mit verschiedenen Datentypen aus einer
   // Ableitungshierarchie? Wenn ich also auf eine A.class "introspecte", kommt
   // der ganze Kram dann mit einem "b" unter Voraussetzung B extends A klar?
}



