package org.colllib.operator;


/**
 * Some useful standard operators.
 * 
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marc.jackisch
 */
public class OperatorCollection {
   private OperatorCollection() {}
   
   /**
    * Return the maximum of two comparable values
    * 
    * @param <T> the data type
    * @return an operator, that returns the bigger value
    */
   public static <T extends Comparable<T>> Operator<T> max() {
      return new Operator<T>() {
         @Override
         public T exec(T t0, T t1) {
            int v = t0.compareTo(t1);
            return v > 0 ? t0 : t1;
         }};
   }

   /**
    * Return the minimum of two comparable values
    * 
    * @param <T> the data type
    * @return an operator, that returns the smaller value
    */
   public static <T extends Comparable<T>> Operator<T> min() {
      return new Operator<T>() {
         @Override
         public T exec(T t0, T t1) {
            int v = t0.compareTo(t1);
            return v < 0 ? t0 : t1;
         }};
   }
   
   
   /**
    * Return the first value always
    * 
    * @param <T> the data type
    * @return an operator, that returns the first operand
    */
   public static <T> Operator<T> first() {
      return new Operator<T>() {
         @Override
         public T exec(T t0, T t1) {
            return t0;
         }};
   }
   
   /**
    * Return the second value always
    * 
    * @param <T> the data type
    * @return an operator, that returns the second operand
    */
   public static <T> Operator<T> second() {
      return new Operator<T>() {
         @Override
         public T exec(T t0, T t1) {
            return t1;
         }};
   }
   
   public static final Operator<Integer> sumInt = new Operator<Integer>() {
      @Override
      public Integer exec(Integer t0, Integer t1) {
         return Integer.valueOf(t0.intValue() + t1.intValue());
      }};

   public static final Operator<Long> sumLong = new Operator<Long>() {
      @Override
      public Long exec(Long t0, Long t1) {
         return Long.valueOf(t0.longValue() + t1.longValue());
      }};
      
   public static final Operator<Float> sumFloat = new Operator<Float>() {
      @Override
      public Float exec(Float t0, Float t1) {
         return Float.valueOf(t0.floatValue() + t1.floatValue());
      }};
      
   public static final Operator<Double> sumDouble = new Operator<Double>() {
      @Override
      public Double exec(Double t0, Double t1) {
         return Double.valueOf(t0.doubleValue() + t1.doubleValue());
      }};
}
