package org.colllib.proxies;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Map;

import org.colllib.transformer.TransformerCollection;
import org.colllib.util.CollectionUtil;

/**
 * The multi call proxy can be used to collect implementations of any interface
 * and perform an interface call on all of those implementations automatically.
 * <p>
 * Originally intented for swing-event multicasting, I find it works for
 * everything... <code>:-)</code>
 * <p>
 * Here's an example for the original intent, i.e. event multicasting:
 * <pre>
 * private MulticallProxy&lt;ChangeListener&gt; changeListeners;
 * ...
 * public void addChangeListener(ChangeListener ch) {
 *   changeListeners.add(ch);
 * }
 *  
 * public void removeChangeListener(ChangeListener ch) {
 *   changeListeners.remove(ch);
 * }
 * 
 * private void fireChange(ChangeEvent e) {
 *   // calls stateChanged on all listeners added:
 *   changeListeners.getProxyInterface().stateChanged(e);
 * }
 * </pre>
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marc.jackisch
 */
public class MulticallProxy<L> extends ArrayList<L> implements InvocationHandler {
   private Class<? super L>  ifcClass;
   private L                 proxy;

   /**
    * Create a new MulticallProxy for a given interface class
    * @param ifcClass the interface class
    */
   public MulticallProxy(Class<? super L> ifcClass) {
      this.ifcClass = ifcClass;
   }
   
   /**
    * Return the usable implementation of the interface class
    * 
    * @return The proxy object to call the interface methods on. If a method
    * thus called has a return value, the <b>last</b> element's return value
    * will be returned by interface calls.
    */
   @SuppressWarnings("unchecked")
   public synchronized L getProxyInterface() {
      if(proxy == null)
         proxy = (L) Proxy.newProxyInstance(
            getClass().getClassLoader(),
            new Class<?>[] { ifcClass  },
            this);
      return proxy;
   }

   /**
    * Handler implementation
    */
   @Override
   public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
      Map<String, Method> methods = CollectionUtil.generateLookupMap(
         CollectionUtil.asList(ifcClass.getDeclaredMethods()),
         TransformerCollection.<Method, String>beanValue("name"));

      if(methods.containsKey(method.getName())) {
         Object lastRes = null;
         for(L l : new ArrayList<L>(this))
            lastRes = methods.get(method.getName()).invoke(l, args);
         return lastRes;
      }

      throw new NoSuchMethodError();
   }
}
