package org.colllib.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.RandomAccess;
import java.util.Set;

import org.colllib.datastruct.AutoInitMap;
import org.colllib.datastruct.ImmutableArrayBackedList;
import org.colllib.datastruct.Pair;
import org.colllib.factories.FactoryCollection;
import org.colllib.filter.Filter;
import org.colllib.filter.FilterCollection;
import org.colllib.generator.Generator;
import org.colllib.operator.Operator;
import org.colllib.operator.OperatorCollection;
import org.colllib.transformer.Transformer;
import org.colllib.transformer.TransformerCollection;


/**
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 *
 * @author marc.jackisch
 */
public class CollectionUtil {
   private CollectionUtil() {}

   /**
    * Inner filtering mechanism
    * @param in       input collection
    * @param filter   the filter
    * @param target   target collection for matching elements
    * @return target, or null if in was null
    */
   private static <T, E extends Collection<T>> E applyFilterTo(Iterable<? extends T> in, Filter<T> filter, E target) {
      if(in == null)
         return null;

      for(T t : in)
         if(filter.matches(t))
            target.add(t);
      return target;
   }

   /**
    * Apply a filter to an input collection, return a list of matching elements
    * @param in      the input collection
    * @param filter  a filter to apply
    * @return a list of matching elements or null, if in was null
    */
   public static <T> ArrayList<T> applyFilter(Iterable<? extends T> in, Filter<T> filter) {
      return applyFilterTo(in, filter, new ArrayList<T>());
   }

   /**
    * Apply a filter to an input array, return a list of matching elements
    * @param in      the input array
    * @param filter  a filter to apply
    * @return a list of matching elements or null, if in was null
    */
   public static <T> ArrayList<T> applyFilter(T[] in, Filter<T> filter) {
      return applyFilterTo(iterableArray(in), filter, new ArrayList<T>());
   }

   /**
    * Iterable wrapper for an old Enumeration
    *
    * @param e the enumeration
    * @return an iterable over the elements of e
    */
   public static <T> Iterable<T> iterableEnum(final Enumeration<T> e) {
      return new Iterable<T>() {
         @Override
         public Iterator<T> iterator() {
            return new Iterator<T>() {
               @Override
               public boolean hasNext() {
                  return e.hasMoreElements();
               }

               @Override
               public T next() {
                  return e.nextElement();
               }

               @Override
               public void remove() {
                  throw new UnsupportedOperationException();
               }
            };
         }
      };
   }

   /**
    * Enumeration wrapper around an iterable.
    * 
    * @param i the iterable
    * @return an enumeration over the elements of i
    */
   public static <T> Enumeration<T> enumIterable(final Iterable<T> i) {
      return new Enumeration<T>() {
         Iterator<T> it = i.iterator();
         @Override
         public boolean hasMoreElements() {
            return it.hasNext();
         }

         @Override
         public T nextElement() {
            return it.next();
         }
      };
   }
   
   /**
    * Iterable wrapper over a basic array
    *
    * @param a an array
    * @return an iterable over the elements of a
    */
   public static <T> Iterable<T> iterableArray(final T[] a) {
      return new Iterable<T>() {
         @Override
         public Iterator<T> iterator() {
            return new Iterator<T>() {
               private int i = 0;
               @Override
               public boolean hasNext() {
                  return i < a.length;
               }

               @Override
               public T next() {
                  try {
                     return a[i++];
                  }
                  catch(IndexOutOfBoundsException e) {
                     throw new NoSuchElementException();
                  }
               }

               @Override
               public void remove() {
                  throw new UnsupportedOperationException();
               }
            };
         }
      };
   }

   /**
    * Iterable wrapper to iterate through two collections parallely
    *
    * @param a the first iterable
    * @param b the second iterable
    * @param pad boolean to determine if the size is calculated by the shorter
    *           or longer collection. If padToMax is set, elements which are not
    *           present in the shorter collection are filled with
    *           <code>null</code>.<br>
    * @return an iterable for object pairs consisting of a and b's elements in
    *         which <code>remove()</code> is not supported.
    */
   public static <A, B> Iterable<Pair<A, B>> pairIterable(Iterable<A> a, Iterable<B> b, final boolean pad) {
      final Iterator<A> itA = a.iterator();
      final Iterator<B> itB = b.iterator();

      return new Iterable<Pair<A, B>>() {
         @Override
         public Iterator<Pair<A, B>> iterator() {
            return new Iterator<Pair<A, B>>() {
               @Override
               public boolean hasNext() {
                  if(pad)
                     return itA.hasNext() || itB.hasNext();
                  else
                     return itA.hasNext() && itB.hasNext();
               }

               @Override
               public Pair<A, B> next() {
                  return new Pair<A, B>(itA.hasNext() ? itA.next() : null, itB.hasNext() ? itB.next() : null);
               }

               @Override
               public void remove() {
                  throw new UnsupportedOperationException();
               }};
         }};
   }

   /**
    * Find first element in a collection that matches a given filter
    *
    * @param in input sequence
    * @param f  the filter
    * @return first index of a matching element or -1
    */
   public static <T> int indexOf(Iterable<T> in, Filter<T> f) {
      int i = 0;
      for(T t : in)
         if(f.matches(t))  return i;
         else              i++;
      return -1;
   }

   /**
    * Find the first element in a collection, that matches a given filter
    *
    * @param in input sequence
    * @param f  the filter
    * @return first matching element or null
    */
   public static <T> T firstElement(Iterable<T> in, Filter<T> f) {
      for(T t : in)
         if(f.matches(t))
            return t;
      return null;
   }

   /**
    * Inner transformation mechanism
    *
    * @param s      source
    * @param t      transformer
    * @param res    destination
    * @return res
    */
   private static <S, D, C extends Collection<D>> C transformTo(Iterable<S> s, Transformer<S, D> t, C res) {
      for(S e : s)
         res.add(t.transform(e));
      return res;
   }

   /**
    * Transform a collection using a given transformer
    *
    * @param <S>  transformation source type
    * @param <D>  transformation destination type
    * @param s    source sequence
    * @param t    transformer
    * @return a list of transformation results
    */
   public static <S, D> ArrayList<D> transform(Iterable<S> s, Transformer<S, D> t) {
      return transformTo(s, t, new ArrayList<D>());
   }

   /**
    * Transform an array using a given transformer
    *
    * @param <S>  transformation source type
    * @param <D>  transformation destination type
    * @param s    source array
    * @param t    transformer
    * @return a list of transformation results
    */
   public static <S, D> ArrayList<D> transform(S[] s, Transformer<S, D> t) {
      return transformTo(iterableArray(s), t, new ArrayList<D>());
   }

   /**
    * Transform a collection using a given transformer
    *
    * @param <S>  transformation source type
    * @param <D>  transformation destination type
    * @param s    source sequence
    * @param t    transformer
    * @return a set of transformation results
    */
   public static <S, D> HashSet<D> transformToSet(Iterable<S> s, Transformer<S, D> t) {
      return transformTo(s, t, new HashSet<D>());
   }

   /**
    * Transform an array using a given transformer
    *
    * @param <S>  transformation source type
    * @param <D>  transformation destination type
    * @param s    source array
    * @param t    transformer
    * @return a set of transformation results
    */
   public static <S, D> HashSet<D> transformToSet(S[] s, Transformer<S, D> t) {
      return transformTo(iterableArray(s), t, new HashSet<D>());
   }

   /**
    * Create a map from an input sequence using a key transformation and a value
    * transformation. For colliding keys, the existing and the new value will
    * be passed through a given operator and the result will be put into the
    * map
    *
    * @param <S>           source type
    * @param <K>           key type
    * @param <V>           value type
    * @param scoll         input sequence
    * @param keyTrans      the key transformer
    * @param valueTrans    the value transformer
    * @param keyCollisionValueOperator operator, which combines an existing
    * and a new value in the map. The arguments are: first the existing value
    * in the map, second the new value
    * @return A map that contains the key transformation results as keys and
    * the value transformation results as values. In case of key collisions,
    * the operator determines the result value
    */
   public static <S, K, V> HashMap<K, V> generateLookupMap(Iterable<S> scoll, Transformer<S, K> keyTrans, Transformer<S, V> valueTrans, Operator<V> keyCollisionValueOperator) {
      HashMap<K, V> res = new HashMap<K, V>();
      for(S s : scoll) {
         K key = keyTrans.transform(s);
         V val = valueTrans.transform(s);
         if(res.containsKey(key))
            val = keyCollisionValueOperator.exec(res.get(key), val);
         res.put(key, val);
      }
      return res;
   }

   /**
    * Create a map from an input sequence using a key transformation and a value
    * transformation
    *
    * @param <S>           source type
    * @param <K>           key type
    * @param <V>           value type
    * @param scoll         input sequence
    * @param keyTrans      the key transformer
    * @param valueTrans    the value transformer
    * @return A map that contains the key transformation results as keys and
    * the value transformation results as values. As the input sequence gets
    * processed completely, the last key "wins" in case of key collisions.
    */
   public static <S, K, V> HashMap<K, V> generateLookupMap(Iterable<S> scoll, Transformer<S, K> keyTrans, Transformer<S, V> valueTrans) {
      return generateLookupMap(scoll, keyTrans, valueTrans, OperatorCollection.<V>second());
   }

   /**
    * As generateLookupMap above, but using the input values for values instead
    * of a transformation result.
    */
   public static <S, K> HashMap<K, S> generateLookupMap(Iterable<S> scoll, Transformer<S, K> keyTrans) {
      return generateLookupMap(scoll, keyTrans, TransformerCollection.<S>id());
   }

   /**
    * Add all elements of a given iterable to a given collection, and return
    * that collection.
    *
    * @param <T>                 element type
    * @param <C>                 collection type
    * @param targetCollection    collection to add the elements to
    * @param elements            the elements
    * @return targetCollection
    */
   public static <T, C extends Collection<? super T>> C addAll(C targetCollection, Iterable<T> elements) {
      for(T t : elements)
         targetCollection.add(t);
      return targetCollection;
   }

   /**
    * Add all elements of a given array to a given collection, and return
    * that collection.
    *
    * @param <T>                 element type
    * @param <C>                 collection type
    * @param targetCollection    collection to add the elements to
    * @param elements            the elements
    * @return targetCollection
    */
   public static <T, C extends Collection<? super T>> C addAll(C targetCollection, T ... elements) {
      return addAll(targetCollection, iterableArray(elements));
   }

   /**
    * Convert an array to an array list
    *
    * @param <T>   element type
    * @param els   the elements
    * @return an array list containing all elements in the same order
    */
   public static <T> ArrayList<T> asList(T ... els) {
      ArrayList<T> res = new ArrayList<T>();
      return addAll(res, iterableArray(els));
   }

   /**
    * Convert an array to a hash set
    *
    * @param <T>   element type
    * @param els   the elements
    * @return a hash set containing all elements
    */
   public static <T> HashSet<T> asSet(T... els) {
      HashSet<T> res = new HashSet<T>();
      return addAll(res, iterableArray(els));
   }

   /**
    * Convert an array to an immutable list without copying it
    *
    * @param <T>   element type
    * @param els   the elements
    * @return an immutable array backed list with the given elements as the
    * backing store
    */
   public static <T> ImmutableArrayBackedList<T> asImmutableList(T ... els) {
      return new ImmutableArrayBackedList<T>(els);
   }

   /**
    * Call to Collections.sort, but returning the original list (for call
    * chaining)
    */
   public static <T extends Comparable<? super T>, L extends List<T>> L sort(L list) {
      Collections.sort(list);
      return list;
   }

   /**
    * Call to Collections.sort, but returning the original list (for call
    * chaining)
    */
   public static <T, L extends List<T>> L sort(L list, Comparator<? super T> c) {
      Collections.sort(list, c);
      return list;
   }

   /**
    * Check if a value is contained in a list of samples
    *
    * @param value   the value to be checked
    * @param samples a list of samples to test if the value is contained
    * @return true if the list contains the value
    */
   public static <T> boolean in(T value, T... samples) {
      for(T t : samples)
         if((value != null && value.equals(t)) || (value == null && t == null))
            return true;
      return false;
   }

   /**
    * Split a given collection by its keys determined using the given
    * transformer. The collection's original order is kept during the split.
    *
    * @param input          the input collection
    * @param keyTransformer the transformer to use for element to key transformation
    * @return a K to ArrayList<V> map
    */
   public static <K, V> AutoInitMap<K, ArrayList<V>> split(Iterable<V> input, Transformer<? super V, K> keyTransformer) {
      AutoInitMap<K, ArrayList<V>> mw = new AutoInitMap<K, ArrayList<V>>(new LinkedHashMap<K, ArrayList<V>>(), FactoryCollection.cloneFactory(new ArrayList<V>()));
      split(input, keyTransformer, mw);
      return mw;
   }

   /**
    * Split a given collection by its keys determined using the given transformer.
    * The resulting associations are stored using the given AutoInitMap.
    *
    * @param input          the input collection
    * @param keyTransformer the transformer to use for element to key transformation
    * @param aim            the target AutoInitMap
    */
   public static <K, V> void split(Iterable<V> input, Transformer<? super V, K> keyTransformer, AutoInitMap<? super K, ? extends Collection<V>> aim) {
      for(V v : input)
         aim.get(keyTransformer.transform(v)).add(v);
   }

   /**
    * Convert key value pairs into a map. Note: due to java language limitations
    * you have to ensure the correct data types in the values array by yourself.
    */
   @SuppressWarnings("unchecked")
   public static <K, V> LinkedHashMap<K, V> asMap(K k1, V v1, Object... values) {
      if(values.length % 2 != 0)
         throw new IllegalArgumentException("values.length must be divisible by 2");
      LinkedHashMap<K, V> r = new LinkedHashMap<K, V>();
      r.put(k1, v1);
      for(int i=0; i<values.length; i+=2)
         r.put((K)values[i], (V)values[i+1]);
      return r;
   }

   /**
    * Convenience method which returns the input list or a mutable empty list.
    *
    * @param l  a list or null
    * @return the given list or an empty {@code ArrayList} if the parameter was null
    */
   public static <T> List<T> emptyListIfNull(List<T> l) {
      if(l == null)
         return new ArrayList<T>();
      return l;
   }

   /**
    * Convenience method which returns the input set or a mutable empty set.
    *
    * @param i  a set or null
    * @return the given se or an empty {@code HashSet} if the parameter was null
    */
   public static <T> Set<T> emptySetIfNull(Set<T> i) {
      if(i == null)
         return new HashSet<T>();
      return i;
   }

   /**
    * Generate a elements into a given collection using an input collection
    * and a given generator
    *
    * @param target the target collection to generate to
    * @param coll input collection
    * @param gen the generator object used to generate the new collection
    * @return the generated collection (== target)
    */
   private static <S, D, C extends Collection<D>> C generate(C target, Iterable<? extends S> coll, Generator<S, D> gen) {
      for(S s : coll)
         addAll(target, gen.generate(s));
      return target;
   }

   /**
    * Generate a new collection by a given collection of elements using
    * a given generator
    *
    * @param coll input collection
    * @param gen the generator object used to generate the new collection
    * @return the generated collection
    */
   public static <S, D> ArrayList<D> generate(Iterable<? extends S> coll, Generator<S, D> gen) {
      return generate(new ArrayList<D>(), coll, gen);
   }

   /**
    * Generate a new set by a given collection of elements using
    * a given generator
    *
    * @param coll input collection
    * @param gen the generator object used to generate the new collection
    * @return the generated set
    */
   public static <S, D> LinkedHashSet<D> generateSet(Iterable<? extends S> coll, Generator<S, D> gen) {
      return generate(new LinkedHashSet<D>(), coll, gen);
   }

   /**
    * Apply an operator on a collection and determine the results
    * in the following way:
    * <ul>
    * <li>if the collection is empty, return null
    * <li>if the collection contains a single element, return this element
    * <li>if the collection contains two or more elements, replace the first
    * element with op.execute(element1, element2), and reiterate
    * </ul>
    *
    * @param l    the iterable input
    * @param oper the operator
    * @return see above
    */
   public static <S> S operate(Iterable<S> l, Operator<S> oper) {
      Iterator<S> i = l.iterator();
      if(!i.hasNext())
         return null;

      S s = i.next();
      while(i.hasNext())
         s = oper.exec(s, i.next());
      return s;
   }


   /**
    * Split a list into chunks of no more than max elements.
    * @param l     the list to split
    * @param max   the maximum number of elements
    * @return      list of sub-lists
    */
   public static <T> ArrayList<List<T>> partitionList(List<T> l, int max) {
      if(max <= 0)
         throw new IllegalArgumentException("max must be > 0");

      ArrayList<List<T>> res = new ArrayList<List<T>>();

      if(l instanceof RandomAccess) {
         for(int start = 0, size = l.size(), end; start < size; ) {
            end = Math.min(start + max, size);
            res.add(l.subList(start, end));
            start = end;
         }
      }
      else {
         ArrayList<T> curr = new ArrayList<T>();
         int cnt = 0;
         for(T t : l) {
            curr.add(t);
            if(++cnt == max) {
               res.add(curr);
               curr = new ArrayList<T>();
               cnt = 0;
            }
         }
         if(!curr.isEmpty())
            res.add(curr);
      }
      return res;
   }


   /**
    * Partition a list of objects by a filter: the resulting pair of lists
    * will contain a list of all matching elements in x, and a list of all
    * non-matching elements in y.
    *
    * @param <T>      the element type
    * @param in       input collection
    * @param splitter the filter used to split the list
    * @return pair of lists of matching and non-matching elements
    */
   public static <T> Pair<ArrayList<T>, ArrayList<T>> partitionByFilter(Iterable<T> in, Filter<T> splitter) {
      Pair<ArrayList<T>, ArrayList<T>> res = new Pair<ArrayList<T>, ArrayList<T>>(new ArrayList<T>(), new ArrayList<T>());
      for(T t : in)
         (splitter.matches(t) ? res.x : res.y).add(t);
      return res;
   }


   /**
    * Remove all given keys from a given map. Returns the modified input
    * map.
    *
    * @param <K>     the key type
    * @param <M>     the map type
    * @param map     the map
    * @param keys    the keys to remove
    * @return the input map
    */
   public static <K, M extends Map<K, ?>> M removeAllKeys(M map, Iterable<K> keys) {
      for(K key : keys)
         map.remove(key);
      return map;
   }


   /**
    * Remove all keys from a given map, which are not in the given collection.
    *
    * @param <K>     the key type
    * @param <M>     the map type
    * @param map     the map
    * @param keys    the keys to retain
    * @return the input map after the keys have been removed
    */
   public static <K, M extends Map<K, ?>> M retainAllKeys(M map, Iterable<K> keys) {
      HashSet<K> keepSet = addAll(new HashSet<K>(), keys);
      for(K key : new ArrayList<K>(map.keySet()))
         if(!keepSet.contains(key))
            map.remove(key);
      return map;
   }


   /**
    * Runs filters over a maps keys and values, retains only the key/value
    * pairs, where both the keyFilter and the valueFilter return true.
    *
    * @param <K>     the key type
    * @param <V>     the value type
    * @param <M>     the map type
    * @param map     the input map
    * @param keyFilter    the key filter
    * @param valueFilter  the value filter
    * @return the filtered input map.
    */
   public static <K, V, M extends Map<K, V>> M filterMapInPlace(M map, Filter<K> keyFilter, Filter<V> valueFilter) {
      for(K key : new ArrayList<K>(map.keySet()))
         if(!keyFilter.matches(key) || !valueFilter.matches(map.get(key)))
            map.remove(key);
      return map;
   }

   /**
    * Like filterMapInPlace above, but returns a copy of the map, leaving the
    * original map unmodified.
    *
    * @param <K>     the key type
    * @param <V>     the value type
    * @param map     the input map
    * @param keyFilter    the key filter
    * @param valueFilter  the value filter
    * @return new map, containing keys and values which match the respective
    * filters.
    */
   public static <K, V> HashMap<K, V> filterMap(Map<K, V> map, Filter<K> keyFilter, Filter<V> valueFilter) {
      HashMap<K, V> t = new HashMap<K, V>(map);
      return filterMapInPlace(t, keyFilter, valueFilter);
   }

   /**
    * Convenience call to filterMap, using only a key filter
    *
    * @param <K>     the key type
    * @param <V>     the value type
    * @param keyFilter    the key filter
    * @return new map, containing keys which match the given filter.
    */
   public static <K, V> HashMap<K, V> filterMapKeys(Map<K, V> map, Filter<K> keyFilter) {
      return filterMap(map, keyFilter, FilterCollection.<V>yes());
   }

   /**
    * Convenience call to filterMap, using only a value filter
    *
    * @param <K>     the key type
    * @param <V>     the value type
    * @param valueFilter  the value filter
    * @return new map, containing values which match the given filter.
    */
   public static <K, V> HashMap<K, V> filterMapValues(Map<K, V> map, Filter<V> valueFilter) {
      return filterMap(map, FilterCollection.<K>yes(), valueFilter);
   }


   /**
    * Transform a map into another map entry-wise by using a key transformer
    * and a value transformer for each key/value pair.
    * <p>
    * The transformation is done in the order of inputMap.entrySet(). There is
    * no collision checking, i.e. if the key transformer results in the same
    * value for multiple keys, the destination map will contain the transformed
    * key/value pair for the last key in that iteration order.
    * 
    * @param <KS>     the source key type
    * @param <VS>     the source value type
    * @param <KD>     the destination key type
    * @param <VD>     the destination value type
    * @param <M>      the result map type
    * @param inputMap         the source map
    * @param outputMap        the destination map
    * @param keyTransformer   transformer used for keys
    * @param valueTransformer transformer used for values
    * @return outputMap with the new element
    */
   public static <KS, KD, VS, VD, M extends Map<KD, VD>> M transformMap(Map<KS, VS> inputMap, M outputMap, Transformer<KS, KD> keyTransformer, Transformer<VS, VD> valueTransformer) {
      for(Map.Entry<KS, VS> e : inputMap.entrySet())
         outputMap.put(
            keyTransformer.transform(e.getKey()),
            valueTransformer.transform(e.getValue()));
      return outputMap;
   }
   
   /**
    * Transform the values of a map using a given value transformer.
    * <p>
    * This method is just a convenience shortcut for
    * <code>transformMap(input, new HashMap&lt;K, D&gt;(), TransformerCollection.&lt;K&gt;id(), valueTransformer)</code>
    * 
    * @see CollectionUtil#transformMap(Map, Map, Transformer, Transformer)
    * 
    * @param input the input map
    * @param valueTransformer the value transformer
    * @return a hash map with the same keys as the input and the transformed
    * values
    */
   public static <K, S, D> HashMap<K, D> transformMapValues(Map<K, S> input, Transformer<S, D> valueTransformer) {
      return transformMap(input, new HashMap<K, D>(), TransformerCollection.<K>id(), valueTransformer);
   }
   
   
   /**
    * Encapsulates the most basic use of an auto init map, which is grouping
    * a collection of objects by a single attribute. The key attribute gets
    * determined by a transformer in this case.
    *
    * @param <E>  the element type
    * @param <K>  the key type
    * @param in   the input collection
    * @param keyTransformer transformer for the E to K transformation
    * @return an AutoInitMap, where each key contains a list of elements in
    * iteration order of the original input collection.
    */
   public static <E, K> AutoInitMap<K,ArrayList<E>> groupBy(Iterable<E> in, Transformer<E, K> keyTransformer) {
      AutoInitMap<K, ArrayList<E>> res = new AutoInitMap<K, ArrayList<E>>(
         new LinkedHashMap<K, ArrayList<E>>(),
         FactoryCollection.cloneFactory(new ArrayList<E>()));
      for(E e : in)
         res.get(keyTransformer.transform(e)).add(e);
      return res;
   }
}

